// IRP Test driver
//
// By Enrico Martignetti


//----------------------------------------------------------------------
//
// Defines
// -------




//----------------------------------------------------------------------
//
// Includes
// --------


extern "C" {
#include <NTDDK.h>
}

// Defs shared with ring 3 components.
//

#include "DrvR3.h"

// Definition of the internal device name and symbolic link. They are 
// derived from DRV_DEVICE_NAME which is the public name defined
// into DrvR3.h and used by client applications.

#define DRV_INT_NAME    L"\\Device\\"DRV_DEVICE_NAME
#define DRV_SYM_LINK    L"\\??\\"DRV_DEVICE_NAME



//----------------------------------------------------------------------
//
// Data Types
// ----------


typedef struct _DEVICE_EXTENSION {
	PDEVICE_OBJECT pDevice;
    KTIMER timer;
    KDPC dpc;
	PIRP pIrp;


} DEVICE_EXTENSION, *PDEVICE_EXTENSION;



//----------------------------------------------------------------------
//
// Forward declarations
// --------------------


static VOID DriverUnload (
		IN PDRIVER_OBJECT	pDriverObject	);

static NTSTATUS CreateDevice (
		IN PDRIVER_OBJECT	pDriverObject,
		OUT PDEVICE_EXTENSION*	ppDevExt	);

static NTSTATUS DeleteDevices(
		IN PDRIVER_OBJECT	pDriverObject	);

static NTSTATUS DispatchCreate (
		IN PDEVICE_OBJECT	pDevObj,
		IN PIRP				pIrp			);

static NTSTATUS DispatchClose (
		IN PDEVICE_OBJECT	pDevObj,
		IN PIRP				pIrp			);

NTSTATUS DispatchRead(
	  IN PDEVICE_OBJECT	pDevObj,
	  IN PIRP			pIrp );

VOID TimerDpc(
    IN PKDPC pDpc,
    IN PVOID pCtx,
    IN PVOID arg1,
    IN PVOID arg2 );



//----------------------------------------------------------------------
//
// Functions
// ---------


//++
// Function:	DriverEntry
//
// Description:
//
//      Announces entry points.
//
//		Creates the device object used to send control codes to this
//		module.
//
// Arguments:
//		pDriverObject - Passed from I/O Manager
//		pRegistryPath - UNICODE_STRING pointer to
//						registry info (service key)
//						for this driver
//
// Return value:
//		NTSTATUS signaling success or failure
//--
extern "C" NTSTATUS DriverEntry (
			IN PDRIVER_OBJECT pDriverObject,
			IN PUNICODE_STRING pRegistryPath	) {
	NTSTATUS status;
	PDEVICE_EXTENSION	pDevExt;

	DbgPrint("\nSkeleton driver, compiled %s %s\n",__DATE__, __TIME__);

	// Announce other driver entry points
	pDriverObject->DriverUnload = DriverUnload;
	pDriverObject->MajorFunction[IRP_MJ_CREATE] =
				DispatchCreate;
	pDriverObject->MajorFunction[IRP_MJ_CLOSE] =
				DispatchClose;
	pDriverObject->MajorFunction[IRP_MJ_READ] = 
				DispatchRead;
	
	// Create the logical device
	status =
		CreateDevice(pDriverObject, &pDevExt);
	if (!NT_SUCCESS(status))
		return status;

	KeInitializeTimer(&pDevExt->timer);
    KeInitializeDpc(&pDevExt->dpc, TimerDpc, pDevExt);

	DbgPrint("\nSkeleton driver succesfully loaded.");

	return STATUS_SUCCESS;
}


//++
// Function:	DriverUnload
//
// Description:
//
//      Realases any resource stored into the device extension,
//      then realases the device object.
//
// Arguments:
//		pDriverObject - Passed from I/O Manager
//
// Return value:
//		None
//--

VOID DriverUnload (
		IN PDRIVER_OBJECT	pDriverObject	) {

    PDEVICE_EXTENSION       lpDevExt;

    // Obtain a pointer to the device extension
    lpDevExt = (PDEVICE_EXTENSION)
        pDriverObject->DeviceObject->DeviceExtension;

	// Delete the logical device.
	DeleteDevices(pDriverObject);

	DbgPrint("\nSkeleton driver unloaded.");
	return;
}


//++
// Function:	CreateDevice
//
// Description:
//		Adds a new device.
//
//		Initializes the device extension.
//
// Arguments:
//		pDriverObject - Passed from I/O Manager
//
//		ppDevExt - pointer to a pointer which will receive the
//			address of the device extension returned by the
//			I/O manager.
//
// Return value:
//		NTSTATUS. Possible failure codes:
//
//			Any failure code returned by IoCreateDevice().
//
//			Any failure code returned by IoCreateSymbolicLink().
//
//--
static NTSTATUS CreateDevice (
		IN PDRIVER_OBJECT	pDriverObject,
		OUT PDEVICE_EXTENSION*	ppDevExt	) 
{

	NTSTATUS status;
	PDEVICE_OBJECT pDevObj;
	PDEVICE_EXTENSION pDevExt;
	UNICODE_STRING uRecName, uSymName;

	RtlInitUnicodeString(&uRecName,
        DRV_INT_NAME);

	// Now create the device
	status =
		IoCreateDevice( pDriverObject,
						sizeof(DEVICE_EXTENSION),
						&uRecName,
						FILE_DEVICE_UNKNOWN,
						0, FALSE,
						&pDevObj );
	if (!NT_SUCCESS(status))
		return status;

	// Initialize the Device Extension
	pDevExt = (PDEVICE_EXTENSION)pDevObj->DeviceExtension;

	pDevExt->pDevice = pDevObj;	// back pointer
	*ppDevExt = pDevExt;

	// Form the symbolic link name
	RtlInitUnicodeString(&uSymName,
        DRV_SYM_LINK);
	status = 
		IoCreateSymbolicLink( &uSymName,
							  &uRecName );
	if (!NT_SUCCESS(status)) {

		// if it fails now, must delete Device object
		IoDeleteDevice( pDevObj );
		return status;
	}

	// Made it
	return STATUS_SUCCESS;
}


//++
// Function:	DeleteDevices
//
// Description:
//		Deletes the device created by this driver.
//
// Arguments:
//		pDriverObject - pointer to driver object
//
// Return value:
//		NTSTATUS - success or failure code
//--

NTSTATUS DeleteDevices(IN PDRIVER_OBJECT pDriverObject) 
{

	PDEVICE_OBJECT	pNextObj;
	UNICODE_STRING  uSymName;

	pNextObj = pDriverObject->DeviceObject;
	if (pNextObj != NULL) {

		// Form the symbolic link which must be deleted
        RtlInitUnicodeString(&uSymName, DRV_SYM_LINK);
		IoDeleteSymbolicLink(&uSymName);

		// Delete the device.
		IoDeleteDevice( pNextObj );
	}
	return STATUS_SUCCESS;

}



//++
// Function:	DispatchClose
//
// Description:
//		Handles call from Win32 CreateHandle request.
//
// Arguments:
//		pDevObj - Passed from I/O Manager
//		pIrp - Passed from I/O Manager
//
// Return value:
//		NTSTATUS - always returns STATUS_SUCCESS;
//--

NTSTATUS DispatchClose (
		IN PDEVICE_OBJECT	pDevObj,
		IN PIRP				pIrp			) {
 
	pIrp->IoStatus.Status = STATUS_SUCCESS;
	pIrp->IoStatus.Information = 0;	// no bytes xfered
	IoCompleteRequest( pIrp, IO_NO_INCREMENT );



	return STATUS_SUCCESS;
}


//++
// Function:	DispatchCreate
//
// Description:
//		Handles call from Win32 CreateFile request.
//
// Arguments:
//		pDevObj - Passed from I/O Manager
//		pIrp - Passed from I/O Manager
//
// Return value:
//		NTSTATUS - always returns STATUS_SUCCESS;
//--


NTSTATUS DispatchCreate (
		IN PDEVICE_OBJECT	pDevObj,
		IN PIRP				pIrp			) {

	pIrp->IoStatus.Status = STATUS_SUCCESS;
	pIrp->IoStatus.Information = 0;	// no bytes xfered
	IoCompleteRequest( pIrp, IO_NO_INCREMENT );


	return STATUS_SUCCESS;
}

//++
// Function:	DispatchRead
//
// Description:
//	Handles call for IRP_MJ_READ
//
// Used to test completion of IRPs
//
// Arguments:
//		pDevObj - Passed from I/O Manager
//		pIrp - Passed from I/O Manager
//
// Return value:
//		NTSTATUS - success or failure code
//--

NTSTATUS DispatchRead(
	  IN PDEVICE_OBJECT	pDevObj,
	  IN PIRP			pIrp )
{
	LARGE_INTEGER complDelay;
	LARGE_INTEGER timeout;
	NTSTATUS status;
	BOOLEAN queued;
	PDEVICE_EXTENSION pDevExt;

	DbgPrint("\nSkelDrv - DispatchRead() called.");
	pDevExt = (PDEVICE_EXTENSION) pDevObj->DeviceExtension;

	// The remainder of this routine contains various calls to
	// IoMarkIrpPending(), IoCompleteRequest(), etc., which can be
	// commented/uncommented, to test the various cases.

	IoMarkIrpPending(pIrp);
	DbgPrint("\nSkelDrv - DispatchRead(): IoMarkIrpPending() called.");

	// The fragment below sets up a timer which calls TimerDpc() after 10''.
	// TimerDpc() calls IoCompleteRequest() and can be used to test 
	// completion occurring after this dispatch routine returned.
	//
	pDevExt->pIrp = pIrp;	// TimerDpc takes pIrp form the device 
							// extension and completes it
	complDelay.QuadPart = - 10 * 10000000; // 10'' delay
    queued = KeSetTimer( &pDevExt->timer, complDelay, &pDevExt->dpc);
    if (TRUE == queued) {
        DbgPrint("\nSkelDrv - DispatchRead(): timer already queued");
	} else {
        DbgPrint("\nSkelDrv - DispatchRead(): completion queued.");
	}

	// Uncomment the fragment for the NTSTATUS we want to return.

	//pIrp->IoStatus.Status = STATUS_SUCCESS;
	//pIrp->IoStatus.Information = 0;
	//IoCompleteRequest(pIrp, IO_NO_INCREMENT);
	//DbgPrint("\nSkelDrv - DispatchRead(): IoCompleteRequest() called.");

	//DbgPrint("\nSkelDrv - DispatchRead(): returning STATUS_SUCCESS");
	//return STATUS_SUCCESS;

	DbgPrint("\nSkelDrv - DispatchRead(): returning STATUS_PENDING");
	return STATUS_PENDING;
}


//++
// Function:	TimerDpc
//
// Description:
//
// Arguments:
//
// Return value:
//--
VOID TimerDpc(
    IN PKDPC pDpc,
    IN PVOID pCtx,
    IN PVOID arg1,
    IN PVOID arg2 )
{
	PDEVICE_EXTENSION pDevExt;
	PIRP pIrp;

	DbgPrint("\nSkelDrv - TimerDpc(): invoked.");
	pDevExt = (PDEVICE_EXTENSION) pCtx;
	pIrp = pDevExt->pIrp;
	DbgPrint("\nSkelDrv - TimerDpc(): calling IoCompleteReqeust().");
	IoCompleteRequest(pIrp, IO_NO_INCREMENT);
	DbgPrint("\nSkelDrv - TimerDpc(): leaving.");
}
